"""Gripper constants and default values."""

LABWARE_GRIP_FORCE = 15.0  # Newtons
GRIPPER_DECK_DROP_OFFSET = [-0.5, -0.5, -1.5]
GRIPPER_MODULE_DROP_OFFSET = [0.0, 0.0, 1.5]
GRIPPER_PADDLE_WIDTH = 36  # mm
